/*
------------------------------------------------------------------------
Low Level code IO code for linux and/or PC

Dunno where this came from but it was significantly modified by me
leon@lrlabs.com

If you know the original author pleas email me so that I can
give him/her credit.

The latest version has only been tested under linux. It used to compile
under dos with djgpp

------------------------------------------------------------------------
------------------------------------------------------------------------
*/
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>

#ifdef linux
 #include <linux/lp.h>
 #include <sys/time.h>
 #include <asm/system.h>
#else
// #include <lp.h>
 #include <sys/time.h>
 #include <sys/system.h>
#endif


extern int debug;			// Deug level
// static unsigned lp_data[3];		// Current levels
extern unsigned char lp_data[];
static unsigned char port;
static int	port_fd;

/* Sometimes this is in /usr/include/linux/lp.h sometimes not */
// #ifdef linux
// struct lp_struct lp_table[] = {
//	{ 0x3bc, 0, 0, LP_INIT_CHAR, LP_INIT_TIME, LP_INIT_WAIT, NULL, NULL, },
//	{ 0x378, 0, 0, LP_INIT_CHAR, LP_INIT_TIME, LP_INIT_WAIT, NULL, NULL, },
//	{ 0x278, 0, 0, LP_INIT_CHAR, LP_INIT_TIME, LP_INIT_WAIT, NULL, NULL, },
// };
// #else
unsigned int LP_B(unsigned char pport) {
	switch ( pport ) {
	case 1:return 0x3bc;break;
	case 2:return 0x378;break;
	case 3:return 0x278;break;
	default :return 0x3bc;break;
	}
}
// #endif

/* 
Base + 0   Data I/O Address    TTL Output Only/(Tristate->Input)
---------------------------------------------------------------------
Bit:  7       6       5       4       3       2       1       0
Pin:  9       8       7       6       5       4       3       2
---------------------------------------------------------------------

Base + 1   Status I/O Address  TTL Input Only
---------------------------------------------------------------------
Bit:  7 Inv   6       5       4       3       2       1       0
Pin:  11      10      12      13      15      NC      NC      NC 
---------------------------------------------------------------------

Base + 2   Control I/O Address Open-Collector Output/(Input:WR=>04h)
---------------------------------------------------------------------
Bit:  7       6       5       4       3 Inv   2       1 Inv   0 Inv
Pin:  NC      NC  NC/(BiDir) Int      17      16      14      1 
---------------------------------------------------------------------

*/

struct {
  unsigned int	offset;
  unsigned char	mask;
} pin_table[26] = {
  { 1, 0x00 },			/* pin 0	Invalid pin #	*/
  { 2, 0x01 },			/* pin 1	(out)	!strobe	*/
  { 0, 0x01 },			/* pin 2	(out)	Data 0	*/
  { 0, 0x02 },			/* pin 3	(out)	Data 1	*/
  { 0, 0x04 },			/* pin 4	(out)	Data 2	*/
  { 0, 0x08 },			/* pin 5	(out)	Data 3	*/
  { 0, 0x10 },			/* pin 6	(out)	Data 4	*/
  { 0, 0x20 },			/* pin 7	(out)	Data 5	*/
  { 0, 0x40 },			/* pin 8	(out)	Data 6	*/
  { 0, 0x80 },			/* pin 9	(out)	Data 7	*/
  { 1, 0x40 },			/* pin 10	(in)	!ack	*/
  { 1, 0x80 },			/* pin 11	(in)	busy	*/
  { 1, 0x20 },			/* pin 12	(in)	Pout	*/
  { 1, 0x10 },			/* pin 13	(in)	Select	*/
  { 2, 0x02 },			/* pin 14	(out)	!feed	*/
  { 1, 0x08 },			/* pin 15	(in)	!error	*/
  { 2, 0x04 },			/* pin 16	(out)	!init	*/
  { 2, 0x08 },			/* pin 17	(out)	!SI	*/
  { 1, 0x00 },			/* pin 18	GND		*/
  { 1, 0x00 },			/* pin 19	GND		*/
  { 1, 0x00 },			/* pin 20	GND		*/
  { 1, 0x00 },			/* pin 21	GND		*/
  { 1, 0x00 },			/* pin 22	GND		*/
  { 1, 0x00 },			/* pin 23	GND		*/
  { 1, 0x00 },			/* pin 24	GND		*/
  { 1, 0x00 },			/* pin 25	GND		*/
};
int	TimeVal;


/* -----------------------------------------------------------------------
   Port control routines -- easier than asm/io.h's routines
   ----------------------------------------------------------------------- */
static inline void outb0 ( char val, short port)
{
#ifdef linux
  ioperm(port, 1, 1);
#endif
  __asm__ volatile ("out%B0 %0,%1" : :"a" (val), "d" (port));
}

static inline unsigned char inb0 (short port)
{
  unsigned int ret;
#ifdef linux
  ioperm(port, 1, 1);  
#endif
  __asm__ volatile ("in%B0 %1,%0" : "=a" (ret) : "d" (port));
  return ret;
}

// ------------------------------------------------------------------------
// Delay for a number of uS -- Works better than usleep for me!
// ------------------------------------------------------------------------
#ifndef linux
void calibrate() {
struct	timeval	tv,tv1;
unsigned int	i,j;

 gettimeofday( &tv, NULL );
 tv1.tv_usec = tv.tv_usec + 100;
 while (tv.tv_usec!=tv1.tv_usec) {
  i++;
  gettimeofday( &tv, NULL );
  gettimeofday( &tv1, NULL );
 }
 i=0;j=0;
 while (tv.tv_usec==tv1.tv_usec) {
  i++;gettimeofday(&tv1, NULL);
 }
 i=0;j=0;tv.tv_usec=tv1.tv_usec;
 while (tv.tv_usec==tv1.tv_usec) {
  i++;if (i==0) j++;
  gettimeofday(&tv1, NULL);
 }
 TimeVal=54644/i;		// Number of us per iteration
 printf("calibrate = %d, %d\n",i,TimeVal);
}
#endif

void waitus( unsigned int us )
{
  struct timeval tv, wt;

#ifdef linux
  gettimeofday( &tv, NULL );
  wt.tv_usec = ( tv.tv_usec + us ) % 1000000;
  wt.tv_sec = tv.tv_sec + ( ( tv.tv_usec + us ) / 1000000);
  /* loop until time is up */
  for (;;) {
    gettimeofday( &tv, NULL );
    if ( tv.tv_sec > wt.tv_sec )
      return;
    if ( tv.tv_sec == wt.tv_sec )
      if ( tv.tv_usec > wt.tv_usec )
        return;
  }
#else
int	i,j;
 i=us;i++;
	while ( i!=0) {
	 for (j=0;j<100;j++) { };
	 i--;
	}
//=us / TimeVal;i++;
//hile (i!=0) {
//i--;
//gettimeofday( &tv, NULL );
//}
#endif
}


// ----------------------------------------------------------------------- 
//   Low Level programmer control
//   pconfig has the pins and polarity of the various control lines
//   pin_table[] has the offset and mask for eack of the DB-25 pins.
// -----------------------------------------------------------------------

/* turn on an output */
int biton( int pin, int pol )
{
  if( pol == 1 ) {
    lp_data[pin_table[pin].offset] |= pin_table[pin].mask;
  } else {
    lp_data[pin_table[pin].offset] &= ~pin_table[pin].mask;
  }
// printf("O(%03x)=%02x,", LP_B( port ) + pin_table[pin].offset, lp_data[pin_table[pin].offset]);
  outb0( lp_data[pin_table[pin].offset], LP_B( port ) + pin_table[pin].offset );

  return 1;
}

/* turn off an output */
int bitoff( int pin, int pol )
{
  if( pol == 1 ) {
    lp_data[pin_table[pin].offset] &= ~pin_table[pin].mask;
  } else {
    lp_data[pin_table[pin].offset] |= pin_table[pin].mask;
  }
// printf("O(%03x)=%02x,", LP_B( port ) + pin_table[pin].offset, lp_data[pin_table[pin].offset]);
  outb0( lp_data[pin_table[pin].offset], LP_B( port ) + pin_table[pin].offset );
  return 1;
}

int inbit(int pin, int pol)
{
  if( pol == 1 ) {
    return( inb0( LP_B( port ) + pin_table[ pin ].offset )
                 & pin_table[ pin ].mask );
  } else {
    return( ~inb0( LP_B( port ) + pin_table[ pin ].offset )
                 & pin_table[ pin ].mask );
  }
}

// -----------------------------------------------------------------------
// Try to initalize the LP port, find it's base address etc. to make sure
//that it is useable.
// -----------------------------------------------------------------------
int init_lpt_port(int pport) {
	char	printer[10];

port=pport;
#ifdef linux

sprintf(printer,"/dev/lp%d", port );
// port_fd=open(printer,O_WRONLY);
// if ( port_fd<0 ) {
//	perror("printer port un init");
//	return 0;
// }
if ( debug ) printf( "I/O base address for %s is 0x%x\n", printer, LP_B( port ) );
// Get permission to access the config register
if ( ioperm( LP_B( port )+0, 3L, 1) <0 ) {
	printf("access to port 0x%x denied\n", LP_B( port )+1 );
	return 0;
}
#else
printf( "I/O base address for LPT%d is 0x%x\n", port, LP_B( port ) );
#endif
// Get the initial states of data and command lines
lp_data[0] = inb0( LP_B( port ) + 0 );
lp_data[1] = inb0( LP_B( port ) + 1 );
lp_data[2] = inb0( LP_B( port ) + 2 );


lp_data[0]=0;


if ( debug ) printf( "data = %02X   status = %02X   cmnd = %02X\n",lp_data[0], lp_data[1], lp_data[2] );
return 1;
}
// -----------------------------------------------------------------------
// Try to initalize the LP port, find it's base address etc. to make sure
//that it is useable.
// -----------------------------------------------------------------------
int init_lpt_port_add(int pport) {
//	char	printer[10];
	int	i;

#ifdef linux

//sprintf(printer,"/dev/lp%d", port );
// port_fd=open(printer,O_WRONLY);
// if ( port_fd<0 ) {
//	perror("printer port un init");
//	return 0;
// }
 for (i=0; i<3; i++) {
  if (pport==LP_B(i)) port=i;
 }
//if ( debug ) printf( "I/O base address for %s is 0x%x\n", printer, LP_B( port ) );
// Get permission to access the config register
if ( ioperm( LP_B( port )+0, 3L, 1) <0 ) {
	printf("access to port 0x%x denied\n", LP_B( port )+1 );
	return 0;
}
#else
printf( "I/O base address for LPT%d is 0x%x\n", port, LP_B( port ) );
#endif
// Get the initial states of data and command lines
lp_data[0] = inb0( LP_B( port ) + 0 );
lp_data[1] = inb0( LP_B( port ) + 1 );
lp_data[2] = inb0( LP_B( port ) + 2 );


lp_data[0]=0;


if ( debug ) printf( "data = %02X   status = %02X   cmnd = %02X\n",lp_data[0], lp_data[1], lp_data[2] );
return 1;
}


int destroy_lpt_port(int i) {

if (i != 0) {
 outb0(LP_B( port )+0, lp_data[0]);
 outb0(LP_B( port )+1, lp_data[1]);
 outb0(LP_B( port )+2, lp_data[2]);
}
if ( debug ) printf( "Port data restored\n" );
#ifdef linux
(void)close(port_fd);
if ( debug ) printf( "Closed lp%d.\n",port);
#endif
return 0;
}
