/* file pgm89.c
 Atmel 89s53, 89s52 programmer
 (C) CDS 2000-2002. All rights reserved.

 Written by Leon Lessing <leon@lrlabs.com>

 compile script:
 gcc -c lowlevel.c
 gcc -c pgmatmel.c
 gcc -o pgmatmel lowlevel.o pgmatmel.o

*/
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <sys/time.h>
#include "lowlevel.h"
#include <ctype.h>

#include <unistd.h>
#include <sys/io.h>


#define	NEWPORT

#ifndef uint
#define uint unsigned int
#endif

typedef int BOOL;
#define TRUE 1
#define FALSE 0


//
// connections
//
// Pin 9 = CPU id bits		(gen 1)		not used
// Pin 8 = RESET		(gen 0)		used
// Pin 7 = Normal Voltage	(sel 1)		used
// Pin 6 = Low voltage		(sel 0)		used
// Pin 5 = VPP on				not used
// Pin 4 = power on				used
// Pin 3 = clk					used
// Pin 2 = data					used
//
// Pin 10= data in loop				used
// Pin 11= s53 datain				used
//
// Pin 14= Sel 2	(Low VPP)		not used
// Pin 15= Sel 3	(Normal VPP)		not used
//
#define	MID	9
#define	RESET	8
#define	NV	7
#define	LV	6
#define	VPROG	5
#define	POWER	4
#define	CLOCK	3
#define	DATA	2
#define	DIN	11
#define DATAIN	10
//
// Serial commands to send to the chip
// Command for s53
#define	SC_PE	0xac53
#define	SC_CE	0xac04
#define	SC_RCM	0x0100
#define	SC_WCM	0x0200
#define	SC_WLB	0xac07

// Command for c8252
#define	SC_RDM	0x0500
#define	SC_WDM	0x0600

int debug=0;
unsigned char	VERSION[7]="1.0.0";
unsigned short	lp_data[3]={0, 0, 0};
unsigned char	buf[20000];
unsigned int	endframe=(((65536/8)>>4)+1);
unsigned char	SerialP;
// unsigned int	endframe=10;
int		PORT=1;		// Port number to use
int		Device=0;	// Default 89s53
unsigned char	DoProtect=0;
char		*fname=NULL;
char		*SerialNo=NULL;



//---------------------------------------------
void vddnormal()	{ bitoff(LV,1);biton (NV,1); }
void vddlow()		{ biton (LV,1);bitoff(NV,1); }
void vddhigh()		{ bitoff(LV,1);bitoff(NV,1); }
//---------------------------------------------
void vppon()	{ biton  (VPROG,1); }
void vppoff()	{ bitoff (VPROG,1); }
//---------------------------------------------
void pwron()	{ biton  (POWER,1); }
void pwroff()	{ bitoff (POWER,1); }
//---------------------------------------------
void idon()	{ biton  (MID,1); }
void idoff()	{ bitoff (MID,1); }
//---------------------------------------------
void resethi()	{ biton  (RESET,0); }
void resetlo()	{ bitoff (RESET,0); }
//---------------------------------------------
void clkhi()	{ biton  (CLOCK,0);if ( (debug & 2) !=0) printf("c=%02x",lp_data[0]); }
void clklo()	{ bitoff (CLOCK,0);if ( (debug & 2) !=0) printf("c=%02x",lp_data[0]); }
//---------------------------------------------
void datahi()	{ biton  (DATA,0);if ( (debug & 2) !=0) printf("d=%02x",lp_data[0]); }
void datalo()	{ bitoff (DATA,0);if ( (debug & 2) !=0) printf("d=%02x",lp_data[0]); }
//---------------------------------------------
/*
int  clkin()	{
 if (inbit(CLKIN,1) !=0) return 1;
  else return 0;
}
*/
//---------------------------------------------
int  datain()	{
 if (inbit(DIN,1) !=0) return 1;
  else return 0;
}
//---------------------------------------------
void TXbits(unsigned int j,char s) {
char     k;
for(k=s-1; k!=-1; k--) {
 if ( ((1 << k) & j)!=0 ) {
  datahi();waitus(10);
  clkhi();waitus(10);
  clklo();      // waitus(30);
  if ( (debug & 4) !=0) printf("1");
 } else {
  datalo();waitus(10);
  clkhi();waitus(10);
  clklo();      // waitus(30);
  if ( (debug & 4) !=0) printf("0");
 }
}
datalo();
}
//---------------------------------------------
void TXaddr(unsigned int i, const unsigned int ormask) {
char     k;
unsigned int	j;
j=i & 0xff;
j|=( (i & 0x1f00) << 3);
j|=ormask;
if (i & 0x2000) {
 j|=0x0400;
}
//printf("%d-%04x ",i,j);
for(k=15; k!=-1; k--) {
 if ( ((1 << k) & j)!=0 ) {
  datahi();waitus(10);
  clkhi();waitus(10);
  clklo();      // waitus(30);
  if ( (debug & 4) !=0) printf("1");
 } else {
  datalo();waitus(10);
  clkhi();waitus(10);
  clklo();      // waitus(30);
  if ( (debug & 4) !=0) printf("0");
 }
}
datalo();
}
//---------------------------------------------
unsigned char RX8bits() {
unsigned char	j;
char		k;
 j=0;
 for(k=7; k!=-1; k--) {
  clkhi();waitus(10);
  if (datain()) {
   j |= (1 << k); 
  }
  clklo();waitus(10);
 }
 return j;
}
//---------------------------------------------
// Used by Hex Loader
//
int ParseLine(const char *line) {
uint	len;
int	addr,type,sum,used;
uint	val;
uint	byte;
const char	*s;
 if (3 != sscanf(line, ":%02X%04X%02X%n", &len, &addr, &type, &used)) {
  printf("Error In Intel hex record\n");return(0);
 }
 // Data type in intel hex file
 if (type == 0x01) return 0;
 if (type || !len) {
  printf("Record lengt error\n");return 0;
 }

 s = &line[used];
 sum = len + addr + type + (addr >> 8);

 // must add a chomp here for dos format files
 if (strchr(s,0x0d)!=NULL) {
  s[strlen(s)-1]=0;
 }

 if(2*(len+1) != strlen(s)) {
  printf("Error Intel record too short\n");return 0;
 }

 for(byte = 0; byte <= len; byte++) {
  if(1 != sscanf(&s[2*byte], "%02X", &val)) {
   printf("Data error in record\n");return 0;
  }
  sum += val;
  if (byte < len ) {
   buf[0]=0;
  }
 }
 if(sum & 0xff) {
  printf("Checksum error\n");return 0;
 }
 return len;
}
//---------------------------------------------
unsigned char writebyte53(unsigned int addr,unsigned char d) {
unsigned int	k;
unsigned char	endit;
 TXaddr(addr,SC_WCM);TXbits(d,8);waitus(100);
 k=0;endit=0;
 while ( (k<50) && (!endit) ) {
  k++;TXaddr(addr,SC_RCM);endit=(d == RX8bits());
  waitus(100);
 }
// printf("t=%d",k);
 return endit;
}
//---------------------------------------------
unsigned char writebyte52(unsigned int addr,unsigned char d) {
unsigned int	k;
unsigned char	endit;
 TXbits(0x40,8);TXbits(addr,16);TXbits(d,8);waitus(100);
 k=0;endit=0;
 while ( (k<50) && (!endit) ) {
  k++;TXbits(0x20,8);TXbits(addr,16);
  endit=(d == RX8bits());
  waitus(100);
 }
// printf("t=%d",k);
 return endit;
}
//---------------------------------------------
unsigned char readbyte53(unsigned int addr) {
 TXaddr(addr,SC_RCM);
 return RX8bits();
}
//---------------------------------------------
unsigned char readbyte52(unsigned int addr) {
 TXbits(0x20,8);TXbits(addr,16);
 return RX8bits();
}
//---------------------------------------------
void Protect52(int LB) {
int i=0;
 waitus(50000);
 TXbits(0xace0,16);
 TXbits(i,16);
 waitus(50000);
 if (LB & 0x01) {		// write lock bit 1
  TXbits(0xace2,16);
  TXbits(i,16);
 }
 waitus(50000);
 if (LB & 0x02) {		// write lock bit 2
  TXbits(0xace1,16);
  TXbits(i,16);
 }
 waitus(50000);
 if (LB & 0x04) {		// write lock bit 3
  TXbits(0xace3,16);
  TXbits(i,16);
 }
 waitus(50000);		// minimum 50ms
 // now verify lock bits
 TXbits(0x24,8);
 i=RX8bits();
 i=RX8bits();
 i= (RX8bits() >> 2) & 0x07;
// i= RX8bits();
 if (i!=LB) {
  printf("Lock bits error want %1X got back %1X\n",LB,i);
 }
}
//---------------------------------------------
void Protect53(int LB) {
int i;
 TXbits(SC_WLB,16);
 i=0;
 TXbits(i,8);
 waitus(50000);		// minimum 10ms
}
//---------------------------------------------
//---------------------------------------------
//---------------------------------------------
//---------------------------------------------
//---------------------------------------------
//
//
int main(int argc, char **argv) {
//unsigned char	temp[100];
unsigned char	pok,SearchState;
unsigned int	i,j,crc1,crc2,Tbytes;
FILE		*fd,*lx_f;
unsigned char	lx_buf[30],*lx_ptr;
unsigned char	kernel_version;

unsigned char	pp;
int		arg;
BOOL		help = FALSE;
BOOL		error = FALSE;

 printf("ATmel programmer version %s\n" \
	"Written by Leon Lessing leon@lrlabs.com\n",VERSION);
 for (arg=1; arg<argc; arg++) {
  if (argv[arg][0] == '-') {
    switch(toupper(argv[arg][1])) {
	case 'P':
		if (sscanf(&argv[arg][2], "%d", &PORT) != 1) error = TRUE;
		if ((PORT < 1) || (PORT > 3)) error = TRUE;
		break;
	case 'D':
		if (sscanf(&argv[arg][2], "%d", &Device) != 1) error = TRUE;
		if ( Device > 1 ) error = TRUE;
		break;
	case 'S':
		SerialNo=&argv[arg][2];
		break;
	case 'W':
		DoProtect=1;
		break;
	case 'V':
		if (sscanf(&argv[arg][2], "%d", &debug) != 1) error = TRUE;
		if ( debug > 9 ) error = TRUE;
		break;
	case 'H':
		help = TRUE;
		break;
	default:
		error = TRUE;
		break;
    }

  } else {	// not parameter assume the filename
    if (fname == NULL) {
	fname = argv[arg];
    } else {
	/* error -- we already found a filename */
	error = TRUE;
    }
  }
  if (error) {
    fprintf(stderr, "Illegal argument: \"%s\"\n", argv[arg]);
    help = TRUE;
    error = FALSE;
  }
 }

 // Check error in acceptable use
 if (help || (fname == NULL)) {
    fprintf(stderr, "Usage:  pgmatmel [options] <filename>\n");
    fprintf(stderr, "\nAvailable options:\n");
    fprintf(stderr, "    -h\t\t\t: show help message\n");
    fprintf(stderr, "    -v<0--9>\t\t: set verbosity level\n");
    fprintf(stderr, "    -p<1 to 3>\t\t: LPT port to use (default = 1)\n");
    fprintf(stderr, "    -d0\t\t\t: at89s53 (default)\n");
    fprintf(stderr, "    -d1\t\t\t: at89s52\n");
    fprintf(stderr, "    -w\t\t\t: protect device\n");
    fprintf(stderr, "    -s<xxxxxxx>\t\t: 7 digit serial number\n");
    return(1);
 }

 // Read the device spec

/*
 memset(&PORT,0,sizeof(PORT));
 strcpy(PORT,ARGV[1]);
 strcpy(fname,ARGV[2]);
 DoProtect=0;
 if (ARGC >= 4) {
  strcpy(temp,ARGV[3]);
  if (temp[0]=='P') {
   DoProtect=1;
   printf("Protecting chip\r\n");
  }
 }
 if (ARGC >= 5) {
  strcpy(SerialNo,ARGV[4]);
 }

*/

#ifndef linux
 calibrate();
#else
 setvbuf(stdout,(char *)NULL,_IONBF,0);
#endif

#ifdef NEWPORT
 lx_f=fopen("/proc/version","r");
 (void)fgets(lx_buf,sizeof(lx_buf),lx_f);
 (void)fclose(lx_f);
 kernel_version=lx_buf[16];
 if ( debug ) printf("Kernel id=2.%c.x\n",kernel_version);
 if (kernel_version=='2'){
    // 2.2 KERNEL
    (void)sprintf(lx_buf,"/proc/parport/%d/hardware",PORT - 1);
 } else {
    // 2.4 KERNEL
    (void)sprintf(lx_buf,"/proc/sys/dev/parport/parport%d/base-addr",PORT - 1);
 }
 lx_f=fopen(lx_buf,"r");
 i=0;
 if (lx_f != NULL){
    (void)fgets(lx_buf,sizeof(lx_buf),lx_f);
    if (kernel_version=='2'){
	// 2.2 kernel
	lx_ptr=strstr(lx_buf,"0x");
	i=(unsigned int)strtol(lx_ptr+2,NULL,16);
    } else {
	// 2.4 kernel
	i=(unsigned int)strtol(lx_buf,NULL,10);
	//printf("pp=(%s) %d\n",lx_buf,i);
    }
    (void)fclose(lx_f);
 } else {
    i=PORT;
    // printf("Sorry port not foundi\n"); return 0;
 }
#endif
 printf("Port=%d addr=%03x\n",PORT,i);
#ifndef NEWPORT
 i=PORT;
#endif
 init_lpt_port_add(i);
 resetlo();clklo();datalo();pwroff();vppoff();idoff();vddnormal();
 printf("Power off Please connect IC, press <return> when ready....\n");
 pp=getchar();

 pwron();waitus(10);resethi();waitus(100);
 resetlo();resethi();waitus(100);

 printf("Powering up\n");
 waitus(100000);

// Enable programming
 printf("Programming enable...");
 i=0;
 switch (Device) {
    case 1:
	    TXbits(SC_PE,16);
	    TXbits(i,8);
	    i=RX8bits();
	    if (i==0x69) printf("OK");
	     else printf("failed");
	    break;
    default:
	    TXbits(SC_PE,16);
	    TXbits(i,8);
	    printf("OK");
	    break;
 }
 waitus(5000);
 printf("\nErase...");

// Erase the chip
 switch (Device) {
    case 1:
	    TXbits(0xac80,16);
	    TXbits(i,16);
	    break;
    default:
	    TXbits(SC_CE,16);
	    TXbits(i,8);
	    break;
 }
 waitus(750000);	// minimum 10ms
 printf("OK\nOpening file %s....",fname);
 crc1=crc2=0;
 fd=fopen(fname,"r");
 if (fd!=NULL) {
  SearchState=0;
  printf("OK\nProgramming.\n0000");
  j=0;pok=1;
  while (!feof(fd) && pok) {
   pp=fgetc(fd);
   switch (SearchState) {
    case 0: if (pp=='S') SearchState++;
	    break;
    case 1: if (pp=='/') SearchState++;
	    else SearchState=0;
	    break;
    case 2: if (pp=='N') SearchState++;
	    else SearchState=0;
	    break;
    case 3: if (pp=='=') SearchState++;
	    else SearchState=0;
	    if (SerialNo!=NULL) {
	     printf("\r\nSerial @%x=",j+1);
	     SerialP=0;
	    } else {
	     SearchState=0;
	    }
	    break;
    case 4: if (SerialNo[SerialP]==0) {
		SearchState=0;
		printf("\r\n");
	    } else {
//		printf(",%c-",pp);
		pp=SerialNo[SerialP++];
		printf("%c",pp);
	    }
	    break;
    default:
	    break;
   }
   crc1+=pp;
   switch (Device) {
	case 1:
		pok=writebyte52(j,pp);
		break;
	default:
		pok=writebyte53(j,pp);
		break;
   }
   j++;
   if (SearchState!=4) {
    if ((j % 8)==0) printf(".");
    if ((j % 128)==0) printf("\r                    \r%04x",j);
   }
  }
  (void)fclose(fd);
  if (pok) printf("OK\n");
   else printf("ERROR @ 0x%04x\n",j-1);
  printf("Verifying....");
  Tbytes=j;
  while (j>0) {
   j--;
   switch (Device){
	case 1:
		crc2+=readbyte52(j);
		break;
	default:
		crc2+=readbyte53(j);
		break;
   }
  }
  if (crc1!=crc2) printf("Error\ncrc1=%x,crc2=%x\n",crc1,crc2);
   else if (debug) printf("OK\ncrc1=%x,crc2=%x\n",crc1,crc2);
    else printf("OK\n");
  if (DoProtect!=0) {
   printf("Protecting Device...");
   // protect the chip
   switch (Device){
    case 1:
	    Protect52(7);
	    break;
    default:
	    Protect53(7);
	    break;
   }
   j=Tbytes;
   printf("OK\nChecking crc.....");
   crc2=0;
   while (j>0) {
    j--;
    switch (Device){
	case 1:
		crc2+=readbyte52(j);
		break;
	default:
		crc2+=readbyte53(j);
		break;
    }
   }
   printf("%x,%x......",crc1,crc2);
   if (crc1==crc2) printf("Error\n");
    else printf("OK\n");
  }
 } else {
  printf("Error\nfile not found.\n");
 }

 printf("Finished\n");
 resetlo();waitus(100);
 pwroff();
 destroy_lpt_port(0);
 return 0;
}

